#
# -*- coding: utf-8 -*-
#
# This file is generated using
#   ./script/generate-geo-data > octodns/record/geo_data.py
# do not modify it directly
#

geo_data = \
    {'AF': {'AO': {'name': 'Angola'},
            'BF': {'name': 'Burkina Faso'},
            'BI': {'name': 'Burundi'},
            'BJ': {'name': 'Benin'},
            'BW': {'name': 'Botswana'},
            'CD': {'name': 'Congo, The Democratic Republic of the'},
            'CF': {'name': 'Central African Republic'},
            'CG': {'name': 'Congo'},
            'CI': {'name': "Côte d'Ivoire"},
            'CM': {'name': 'Cameroon'},
            'CV': {'name': 'Cabo Verde'},
            'DJ': {'name': 'Djibouti'},
            'DZ': {'name': 'Algeria'},
            'EG': {'name': 'Egypt'},
            'EH': {'name': 'Western Sahara'},
            'ER': {'name': 'Eritrea'},
            'ET': {'name': 'Ethiopia'},
            'GA': {'name': 'Gabon'},
            'GH': {'name': 'Ghana'},
            'GM': {'name': 'Gambia'},
            'GN': {'name': 'Guinea'},
            'GQ': {'name': 'Equatorial Guinea'},
            'GW': {'name': 'Guinea-Bissau'},
            'KE': {'name': 'Kenya'},
            'KM': {'name': 'Comoros'},
            'LR': {'name': 'Liberia'},
            'LS': {'name': 'Lesotho'},
            'LY': {'name': 'Libya'},
            'MA': {'name': 'Morocco'},
            'MG': {'name': 'Madagascar'},
            'ML': {'name': 'Mali'},
            'MR': {'name': 'Mauritania'},
            'MU': {'name': 'Mauritius'},
            'MW': {'name': 'Malawi'},
            'MZ': {'name': 'Mozambique'},
            'NA': {'name': 'Namibia'},
            'NE': {'name': 'Niger'},
            'NG': {'name': 'Nigeria'},
            'RE': {'name': 'Réunion'},
            'RW': {'name': 'Rwanda'},
            'SC': {'name': 'Seychelles'},
            'SD': {'name': 'Sudan'},
            'SH': {'name': 'Saint Helena, Ascension and Tristan da Cunha'},
            'SL': {'name': 'Sierra Leone'},
            'SN': {'name': 'Senegal'},
            'SO': {'name': 'Somalia'},
            'SS': {'name': 'South Sudan'},
            'ST': {'name': 'Sao Tome and Principe'},
            'SZ': {'name': 'Swaziland'},
            'TD': {'name': 'Chad'},
            'TG': {'name': 'Togo'},
            'TN': {'name': 'Tunisia'},
            'TZ': {'name': 'Tanzania, United Republic of'},
            'UG': {'name': 'Uganda'},
            'YT': {'name': 'Mayotte'},
            'ZA': {'name': 'South Africa'},
            'ZM': {'name': 'Zambia'},
            'ZW': {'name': 'Zimbabwe'}},
     'AN': {'AQ': {'name': 'Antarctica'},
            'BV': {'name': 'Bouvet Island'},
            'HM': {'name': 'Heard Island and McDonald Islands'},
            'TF': {'name': 'French Southern Territories'}},
     'AS': {'AE': {'name': 'United Arab Emirates'},
            'AF': {'name': 'Afghanistan'},
            'AM': {'name': 'Armenia'},
            'AZ': {'name': 'Azerbaijan'},
            'BD': {'name': 'Bangladesh'},
            'BH': {'name': 'Bahrain'},
            'BN': {'name': 'Brunei Darussalam'},
            'BT': {'name': 'Bhutan'},
            'CC': {'name': 'Cocos (Keeling) Islands'},
            'CN': {'name': 'China'},
            'CX': {'name': 'Christmas Island'},
            'CY': {'name': 'Cyprus'},
            'GE': {'name': 'Georgia'},
            'HK': {'name': 'Hong Kong'},
            'ID': {'name': 'Indonesia'},
            'IL': {'name': 'Israel'},
            'IN': {'name': 'India'},
            'IO': {'name': 'British Indian Ocean Territory'},
            'IQ': {'name': 'Iraq'},
            'IR': {'name': 'Iran, Islamic Republic of'},
            'JO': {'name': 'Jordan'},
            'JP': {'name': 'Japan'},
            'KG': {'name': 'Kyrgyzstan'},
            'KH': {'name': 'Cambodia'},
            'KP': {'name': "Korea, Democratic People's Republic of"},
            'KR': {'name': 'Korea, Republic of'},
            'KW': {'name': 'Kuwait'},
            'KZ': {'name': 'Kazakhstan'},
            'LA': {'name': "Lao People's Democratic Republic"},
            'LB': {'name': 'Lebanon'},
            'LK': {'name': 'Sri Lanka'},
            'MM': {'name': 'Myanmar'},
            'MN': {'name': 'Mongolia'},
            'MO': {'name': 'Macao'},
            'MV': {'name': 'Maldives'},
            'MY': {'name': 'Malaysia'},
            'NP': {'name': 'Nepal'},
            'OM': {'name': 'Oman'},
            'PH': {'name': 'Philippines'},
            'PK': {'name': 'Pakistan'},
            'PS': {'name': 'Palestine, State of'},
            'QA': {'name': 'Qatar'},
            'SA': {'name': 'Saudi Arabia'},
            'SG': {'name': 'Singapore'},
            'SY': {'name': 'Syrian Arab Republic'},
            'TH': {'name': 'Thailand'},
            'TJ': {'name': 'Tajikistan'},
            'TM': {'name': 'Turkmenistan'},
            'TR': {'name': 'Turkey'},
            'TW': {'name': 'Taiwan, Province of China'},
            'UZ': {'name': 'Uzbekistan'},
            'VN': {'name': 'Viet Nam'},
            'YE': {'name': 'Yemen'}},
     'EU': {'AD': {'name': 'Andorra'},
            'AL': {'name': 'Albania'},
            'AT': {'name': 'Austria'},
            'AX': {'name': 'Åland Islands'},
            'BA': {'name': 'Bosnia and Herzegovina'},
            'BE': {'name': 'Belgium'},
            'BG': {'name': 'Bulgaria'},
            'BY': {'name': 'Belarus'},
            'CH': {'name': 'Switzerland'},
            'CZ': {'name': 'Czechia'},
            'DE': {'name': 'Germany'},
            'DK': {'name': 'Denmark'},
            'EE': {'name': 'Estonia'},
            'ES': {'name': 'Spain'},
            'FI': {'name': 'Finland'},
            'FO': {'name': 'Faroe Islands'},
            'FR': {'name': 'France'},
            'GB': {'name': 'United Kingdom'},
            'GG': {'name': 'Guernsey'},
            'GI': {'name': 'Gibraltar'},
            'GR': {'name': 'Greece'},
            'HR': {'name': 'Croatia'},
            'HU': {'name': 'Hungary'},
            'IE': {'name': 'Ireland'},
            'IM': {'name': 'Isle of Man'},
            'IS': {'name': 'Iceland'},
            'IT': {'name': 'Italy'},
            'JE': {'name': 'Jersey'},
            'LI': {'name': 'Liechtenstein'},
            'LT': {'name': 'Lithuania'},
            'LU': {'name': 'Luxembourg'},
            'LV': {'name': 'Latvia'},
            'MC': {'name': 'Monaco'},
            'MD': {'name': 'Moldova, Republic of'},
            'ME': {'name': 'Montenegro'},
            'MK': {'name': 'Macedonia, Republic of'},
            'MT': {'name': 'Malta'},
            'NL': {'name': 'Netherlands'},
            'NO': {'name': 'Norway'},
            'PL': {'name': 'Poland'},
            'PT': {'name': 'Portugal'},
            'RO': {'name': 'Romania'},
            'RS': {'name': 'Serbia'},
            'RU': {'name': 'Russian Federation'},
            'SE': {'name': 'Sweden'},
            'SI': {'name': 'Slovenia'},
            'SJ': {'name': 'Svalbard and Jan Mayen'},
            'SK': {'name': 'Slovakia'},
            'SM': {'name': 'San Marino'},
            'UA': {'name': 'Ukraine'},
            'VA': {'name': 'Holy See (Vatican City State)'}},
     'ID': {'TL': {'name': 'Timor-Leste'}},
     'NA': {'AG': {'name': 'Antigua and Barbuda'},
            'AI': {'name': 'Anguilla'},
            'AW': {'name': 'Aruba'},
            'BB': {'name': 'Barbados'},
            'BL': {'name': 'Saint Barthélemy'},
            'BM': {'name': 'Bermuda'},
            'BQ': {'name': 'Bonaire, Sint Eustatius and Saba'},
            'BS': {'name': 'Bahamas'},
            'BZ': {'name': 'Belize'},
            'CA': {'name': 'Canada'},
            'CR': {'name': 'Costa Rica'},
            'CU': {'name': 'Cuba'},
            'CW': {'name': 'Curaçao'},
            'DM': {'name': 'Dominica'},
            'DO': {'name': 'Dominican Republic'},
            'GD': {'name': 'Grenada'},
            'GL': {'name': 'Greenland'},
            'GP': {'name': 'Guadeloupe'},
            'GT': {'name': 'Guatemala'},
            'HN': {'name': 'Honduras'},
            'HT': {'name': 'Haiti'},
            'JM': {'name': 'Jamaica'},
            'KN': {'name': 'Saint Kitts and Nevis'},
            'KY': {'name': 'Cayman Islands'},
            'LC': {'name': 'Saint Lucia'},
            'MF': {'name': 'Saint Martin (French part)'},
            'MQ': {'name': 'Martinique'},
            'MS': {'name': 'Montserrat'},
            'MX': {'name': 'Mexico'},
            'NI': {'name': 'Nicaragua'},
            'PA': {'name': 'Panama'},
            'PM': {'name': 'Saint Pierre and Miquelon'},
            'PR': {'name': 'Puerto Rico'},
            'SV': {'name': 'El Salvador'},
            'SX': {'name': 'Sint Maarten (Dutch part)'},
            'TC': {'name': 'Turks and Caicos Islands'},
            'TT': {'name': 'Trinidad and Tobago'},
            'US': {'name': 'United States',
                   'provinces': {'AK': {'name': 'Alaska'},
                                 'AL': {'name': 'Alabama'},
                                 'AR': {'name': 'Arkansas'},
                                 'AS': {'name': 'American Samoa'},
                                 'AZ': {'name': 'Arizona'},
                                 'CA': {'name': 'California'},
                                 'CO': {'name': 'Colorado'},
                                 'CT': {'name': 'Connecticut'},
                                 'DC': {'name': 'District of Columbia'},
                                 'DE': {'name': 'Delaware'},
                                 'FL': {'name': 'Florida'},
                                 'GA': {'name': 'Georgia'},
                                 'GU': {'name': 'Guam'},
                                 'HI': {'name': 'Hawaii'},
                                 'IA': {'name': 'Iowa'},
                                 'ID': {'name': 'Idaho'},
                                 'IL': {'name': 'Illinois'},
                                 'IN': {'name': 'Indiana'},
                                 'KS': {'name': 'Kansas'},
                                 'KY': {'name': 'Kentucky'},
                                 'LA': {'name': 'Louisiana'},
                                 'MA': {'name': 'Massachusetts'},
                                 'MD': {'name': 'Maryland'},
                                 'ME': {'name': 'Maine'},
                                 'MI': {'name': 'Michigan'},
                                 'MN': {'name': 'Minnesota'},
                                 'MO': {'name': 'Missouri'},
                                 'MP': {'name': 'Northern Mariana Islands'},
                                 'MS': {'name': 'Mississippi'},
                                 'MT': {'name': 'Montana'},
                                 'NC': {'name': 'North Carolina'},
                                 'ND': {'name': 'North Dakota'},
                                 'NE': {'name': 'Nebraska'},
                                 'NH': {'name': 'New Hampshire'},
                                 'NJ': {'name': 'New Jersey'},
                                 'NM': {'name': 'New Mexico'},
                                 'NV': {'name': 'Nevada'},
                                 'NY': {'name': 'New York'},
                                 'OH': {'name': 'Ohio'},
                                 'OK': {'name': 'Oklahoma'},
                                 'OR': {'name': 'Oregon'},
                                 'PA': {'name': 'Pennsylvania'},
                                 'PR': {'name': 'Puerto Rico'},
                                 'RI': {'name': 'Rhode Island'},
                                 'SC': {'name': 'South Carolina'},
                                 'SD': {'name': 'South Dakota'},
                                 'TN': {'name': 'Tennessee'},
                                 'TX': {'name': 'Texas'},
                                 'UM': {'name': 'United States Minor Outlying '
                                                'Islands'},
                                 'UT': {'name': 'Utah'},
                                 'VA': {'name': 'Virginia'},
                                 'VI': {'name': 'Virgin Islands'},
                                 'VT': {'name': 'Vermont'},
                                 'WA': {'name': 'Washington'},
                                 'WI': {'name': 'Wisconsin'},
                                 'WV': {'name': 'West Virginia'},
                                 'WY': {'name': 'Wyoming'}}},
            'VC': {'name': 'Saint Vincent and the Grenadines'},
            'VG': {'name': 'Virgin Islands, British'},
            'VI': {'name': 'Virgin Islands, U.S.'}},
     'OC': {'AS': {'name': 'American Samoa'},
            'AU': {'name': 'Australia'},
            'CK': {'name': 'Cook Islands'},
            'FJ': {'name': 'Fiji'},
            'FM': {'name': 'Micronesia, Federated States of'},
            'GU': {'name': 'Guam'},
            'KI': {'name': 'Kiribati'},
            'MH': {'name': 'Marshall Islands'},
            'MP': {'name': 'Northern Mariana Islands'},
            'NC': {'name': 'New Caledonia'},
            'NF': {'name': 'Norfolk Island'},
            'NR': {'name': 'Nauru'},
            'NU': {'name': 'Niue'},
            'NZ': {'name': 'New Zealand'},
            'PF': {'name': 'French Polynesia'},
            'PG': {'name': 'Papua New Guinea'},
            'PN': {'name': 'Pitcairn'},
            'PW': {'name': 'Palau'},
            'SB': {'name': 'Solomon Islands'},
            'TK': {'name': 'Tokelau'},
            'TO': {'name': 'Tonga'},
            'TV': {'name': 'Tuvalu'},
            'UM': {'name': 'United States Minor Outlying Islands'},
            'VU': {'name': 'Vanuatu'},
            'WF': {'name': 'Wallis and Futuna'},
            'WS': {'name': 'Samoa'}},
     'SA': {'AR': {'name': 'Argentina'},
            'BO': {'name': 'Bolivia, Plurinational State of'},
            'BR': {'name': 'Brazil'},
            'CL': {'name': 'Chile'},
            'CO': {'name': 'Colombia'},
            'EC': {'name': 'Ecuador'},
            'FK': {'name': 'Falkland Islands (Malvinas)'},
            'GF': {'name': 'French Guiana'},
            'GS': {'name': 'South Georgia and the South Sandwich Islands'},
            'GY': {'name': 'Guyana'},
            'PE': {'name': 'Peru'},
            'PY': {'name': 'Paraguay'},
            'SR': {'name': 'Suriname'},
            'UY': {'name': 'Uruguay'},
            'VE': {'name': 'Venezuela, Bolivarian Republic of'}}}
